"""Extended Textual ContentSwitcher widget."""

from typing import Any
from typing import Iterable

from textual import events
from textual import widgets
from textual.widget import Widget

from .._extensions import Callback
from .._extensions import WidgetExtension


class ContentSwitcher(widgets.ContentSwitcher, WidgetExtension):
    """An extended widget for switching between different children.

    Note:
        All child widgets that are to be switched between need a unique ID.
        Children that have no ID will be hidden and ignored.
    """

    def __init__(
        self,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        initial: str | None = None,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize the content switching widget.

        Args:
            *children: The widgets to switch between.
            name: The name of the content switcher.
            id: The ID of the content switcher in the DOM.
            classes: The CSS classes of the content switcher.
            disabled: Whether the content switcher is disabled or not.
            initial: The ID of the initial widget to show, ``None`` or empty string for the first tab.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.

        Note:
            If `initial` is not supplied, no children will be shown to start with.
        """
        super().__init__(*children, name=name, id=id, classes=classes, disabled=disabled, initial=initial)
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )

    def next_child(self) -> str | None:
        """Find the next child available in the content list.

        Returns:
            The id of the next child if not currently displaying last, None otherwise.
        """
        next_child = None
        children = [child.id for child in self.children if child.id]
        if self.current != children[-1]:
            current_index = children.index(self.current)
            next_child = children[current_index + 1]
        return next_child

    def previous_child(self) -> str | None:
        """Find the previous child available in the content list.

        Returns:
            The id of the previous child if not currently displaying first, None otherwise.
        """
        previous_child = None
        children = [child.id for child in self.children if child.id]
        if self.current != children[0]:
            current_index = children.index(self.current)
            previous_child = children[current_index - 1]
        return previous_child

    def switch_to_next(self) -> str | None:
        """Find and switch to the last child available in the content list.

        Returns:
            The id of the new child if not currently displaying last, None otherwise.
        """
        next_child = self.next_child()
        if next_child:
            self.current = next_child
        return next_child

    def switch_to_previous(self) -> str | None:
        """Find and switch to the previous child available in the content list.

        Returns:
            The id of the new child if not currently displaying first, None otherwise.
        """
        previous_child = self.previous_child()
        if previous_child:
            self.current = previous_child
        return previous_child



"""Extended Textual Loading Indicator widget."""

from time import time
from typing import Any
from typing import Iterable

from textual import events
from textual import widgets
from textual.widget import Widget

from .._extensions import Callback
from .._extensions import WidgetExtension


class LoadingIndicator(widgets.LoadingIndicator, WidgetExtension):
    """An extended display for an animated loading indicator."""

    def __init__(
        self,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize LoadingIndicator with support for extension arguments.

        Args:
            *children: Child widgets.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )
        # Textual-2912: possibility of _start_time reference in render before mount due to not being declared in init.
        self._start_time = time()



"""Extended Textual tooltip widget."""

from textual import widgets

from .._extensions import StaticFactory


class Tooltip(StaticFactory(widgets.Tooltip)):
    """An extended widget for displaying short tip messages."""



"""Extended Textual TextArea widget."""

from typing import Any
from typing import Iterable

from rich.console import RenderableType
from textual import events
from textual import widgets
from typing_extensions import Literal

from .._extensions import Callback
from .._extensions import WidgetExtension


class TextArea(widgets.TextArea, WidgetExtension):
    """An extended multi-line text input widget."""

    def __init__(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        text: str = "",
        *,
        language: str | None = None,
        theme: str = "css",
        soft_wrap: bool = True,
        tab_behavior: Literal["focus", "indent"] = "focus",
        read_only: bool = False,
        show_line_numbers: bool = False,
        max_checkpoints: int = 50,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        tooltip: RenderableType | None = None,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize the text area widget.

        Args:
            text: The initial text to load into the TextArea.
            language: The language to use.
            theme: The theme to use.
            soft_wrap: Enable soft wrapping.
            tab_behavior: If 'focus', pressing tab will switch focus. If 'indent', pressing tab will insert a tab.
            read_only: Enable read-only mode. This prevents edits using the keyboard.
            show_line_numbers: Show line numbers on the left edge.
            max_checkpoints: The maximum number of undo history checkpoints to retain.
            name: The name of the collapsible.
            id: The ID of the widget in the DOM.
            classes: The CSS classes of the widget.
            disabled: Whether the widget is disabled or not.
            tooltip: Optional tooltip.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            text=text,
            language=language,
            theme=theme,
            soft_wrap=soft_wrap,
            tab_behavior=tab_behavior,
            read_only=read_only,
            show_line_numbers=show_line_numbers,
            max_checkpoints=max_checkpoints,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            tooltip=tooltip,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""List item that acts as an unselectable header between groups."""

from typing import Any
from typing import Iterable

from textual import events
from textual.widget import Widget

from ._extensions import Callback
from ._list_item import ListItem


class ListItemHeader(ListItem):
    """A widget that is an item within a ListView that acts as an unselectable header between list items."""

    DEFAULT_CSS = """
    ListItemHeader>Widget {
        color: $text;
        text-style: bold;
    }
    ListItemHeader:hover>Widget {
        color: $text;
        text-style: bold;
    }
    """

    def __init__(
        self,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        disable_click: bool = True,
        data: Any = None,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize a ListItemHeader with extension arguments.

        Args:
            *children: Child widgets.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
            disable_click: Whether clicking the widget it disabled.
            data: Optional data associated with the list item.
                If no child is provided for display, the data will be searched for a "label" key to use in a Label.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            data=data,
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )
        self.disable_click = disable_click

    async def _on_click(self, event: events.Click) -> None:
        """Disable clicking header items."""
        if self.disable_click:
            event.stop()
            event.prevent_default()



"""Collection of new, and extended, Textual widgets.

This interface file must be kept in sync with the __init__.py's `__all__` list.
This file is used by text editors and type checkers to see the lazily loaded classes.
"""

from textual.widget import Widget as Widget

from textology.textual_utils import textual_version as textual_version

from ._button import Button as Button
from ._extensions import Callback as Callback
from ._extensions import Clickable as Clickable
from ._extensions import Static as Static
from ._extensions import Widget as Widget
from ._extensions import WidgetExtension as WidgetExtension
from ._extensions import walk_all_children as walk_all_children
from ._horizontal_menus import HorizontalMenus as HorizontalMenus
from ._list_item import ListItem as ListItem
from ._list_item_header import ListItemHeader as ListItemHeader
from ._list_item_meta import ListItemMeta as ListItemMeta
from ._list_view import ListView as ListView
from ._location import Location as Location
from ._modal_dialog import ModalDialog as ModalDialog
from ._multi_select import MultiSelect as MultiSelect
from ._page_container import PageContainer as PageContainer
from ._popup_text import PopupText as PopupText
from ._store import Store as Store
from ._text import Text as Text
from ._textual._checkbox import Checkbox as Checkbox
from ._textual._collapsible import Collapsible as Collapsible
from ._textual._containers import Center as Center
from ._textual._containers import Container as Container
from ._textual._containers import Grid as Grid
from ._textual._containers import Horizontal as Horizontal
from ._textual._containers import HorizontalScroll as HorizontalScroll
from ._textual._containers import Middle as Middle
from ._textual._containers import ScrollableContainer as ScrollableContainer
from ._textual._containers import Vertical as Vertical
from ._textual._containers import VerticalScroll as VerticalScroll
from ._textual._content_switcher import ContentSwitcher as ContentSwitcher
from ._textual._data_table import DataTable as DataTable
from ._textual._digits import Digits as Digits
from ._textual._directory_tree import DirectoryTree as DirectoryTree
from ._textual._footer import Footer as Footer
from ._textual._header import Header as Header
from ._textual._label import Label as Label
from ._textual._loading_indicator import LoadingIndicator as LoadingIndicator
from ._textual._log import Log as Log
from ._textual._markdown import Markdown as Markdown
from ._textual._markdown import MarkdownViewer as MarkdownViewer
from ._textual._option_list import OptionList as OptionList
from ._textual._pretty import Pretty as Pretty
from ._textual._progress_bar import ProgressBar as ProgressBar
from ._textual._radio_button import RadioButton as RadioButton
from ._textual._radio_set import RadioSet as RadioSet
from ._textual._rich_log import RichLog as RichLog
from ._textual._rule import Rule as Rule
from ._textual._select import Select as Select
from ._textual._selection_list import SelectionList as SelectionList
from ._textual._sparkline import Sparkline as Sparkline
from ._textual._switch import Switch as Switch
from ._textual._tabbed_content import TabbedContent as TabbedContent
from ._textual._tabbed_content import TabPane as TabPane
from ._textual._tabs import Tab as Tab
from ._textual._tabs import Tabs as Tabs
from ._textual._text_area import TextArea as TextArea
from ._textual._text_input import TextInput as TextInput
from ._textual._tooltip import Tooltip as Tooltip
from ._textual._tree import Tree as Tree



"""Extended Textual Input widget."""

from typing import Any
from typing import Iterable

from rich.highlighter import Highlighter
from textual import events
from textual import widgets
from textual.suggester import Suggester
from textual.types import InputValidationOn
from textual.validation import Validator
from typing_extensions import Literal

from .._extensions import Callback
from .._extensions import WidgetExtension

InputType = Literal["integer", "number", "text"]


class TextInput(widgets.Input, WidgetExtension):
    """An extended text input widget.

    Renamed from "Input" to avoid conflicts with callback dependency types.
    """

    def __init__(  # pylint: disable=too-many-arguments,too-many-locals
        self,
        value: str | None = None,
        placeholder: str = "",
        highlighter: Highlighter | None = None,
        password: bool = False,
        *,
        restrict: str | None = None,
        type: InputType = "text",
        max_length: int = 0,
        suggester: Suggester | None = None,
        validators: Validator | Iterable[Validator] | None = None,
        validate_on: Iterable[InputValidationOn] | None = None,
        valid_empty: bool = False,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize the text input widget.

        Args:
            value: An optional default value for the input.
            placeholder: Optional placeholder text for the input.
            highlighter: An optional highlighter for the input.
            password: Flag to say if the field should obfuscate its content.
            restrict: A regex to restrict character inputs.
            type: The type of the input.
            max_length: The maximum length of the input, or 0 for no maximum length.
            suggester: Suggester associated with this input instance.
            validators: An iterable of validators that the Input value will be checked against.
            validate_on: Zero or more of the values "blur", "changed", and "submitted",
                which determine when to do input validation. The default is to do
                validation for all messages.
            valid_empty: Empty values are valid.
            name: Optional name for the input widget.
            id: Optional ID for the widget.
            classes: Optional initial classes for the widget.
            disabled: Whether the input is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            value=value,
            placeholder=placeholder,
            highlighter=highlighter,
            password=password,
            restrict=restrict,
            type=type,
            max_length=max_length,
            suggester=suggester,
            validators=validators,
            validate_on=validate_on,
            valid_empty=valid_empty,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Extended Textual container widgets."""

from __future__ import annotations

from textual import containers

from .._extensions import WidgetFactory


class Center(WidgetFactory(containers.Center)):
    """An extended container which aligns children on the X axis."""


class Container(WidgetFactory(containers.Container)):
    """An extended, simple container widget, with vertical layout."""


class Grid(WidgetFactory(containers.Grid)):
    """An extended container with grid layout."""


class Horizontal(WidgetFactory(containers.Horizontal)):
    """An extended container with horizontal layout and no scrollbars."""


class HorizontalScroll(WidgetFactory(containers.HorizontalScroll)):
    """An extended container with horizontal layout and an automatic scrollbar on the Y axis."""


class Middle(WidgetFactory(containers.Middle)):
    """An extended container which aligns children on the Y axis."""


class ScrollableContainer(WidgetFactory(containers.ScrollableContainer)):
    """An extended scrollable container with vertical layout, and auto scrollbars on both axis."""


class Vertical(WidgetFactory(containers.Vertical)):
    """An extended container with vertical layout and no scrollbars."""


class VerticalScroll(WidgetFactory(containers.VerticalScroll)):
    """An extended container with vertical layout and an automatic scrollbar on the Y axis."""



"""Dash compatibility layer for creating Textual applications in "Plotly Dash" code style.

The goal of this library is to provide a familiar interface for developers who have worked with Plotly Dash
to make UI web applications, due to usage of similar principles. This can be used for most applications,
or to help transition developers to the style used in Textology for "observation pattern" type applications.

While this library provides compatibility across frameworks in style where possible, it is not a guarantee or promise
for any specific features or functionality. If a feature in Textology extended applications has an equivalent in Dash,
it will be included here, but this is not Dash. In order to access the full range of Textology and Textual features,
developers may need to leverage the native classes and designs from the base libraries.
"""

import logging
from types import ModuleType
from typing import Callable

from textual.app import CSSPathType
from textual.driver import Driver
from textual.widget import Widget

from .apps import ExtendedApp

# Compatibility aliases for Dash.
# Allow importing other modules for direct access through this module for simplicity. pylint: disable=unused-import
from .observers import Dependency
from .observers import Modified
from .observers import NoUpdate
from .observers import Published
from .observers import Raised
from .observers import Select as State
from .observers import SupportsID
from .observers import Update as Output
from .observers import when as callback
from .pages import _GLOBAL_PAGE_MAP as page_registry
from .pages import Page

InputType = Modified | Published | Raised


class DashCompatApp(ExtendedApp):
    """Application capable of performing automatic input/output callbacks on reactive widget property updates.

    Compatibility alias for ExtendedApp class.

    Additional functionality (provided by ExtendedApp and wrapped with Dash compat):
        - Automatic input/output callbacks for managing application state via ".callback()" registration.
        - URL based multi-page content via ".register_page()".
        - URL routing for resource requests within the application via ".location.get()".
        - URL history for navigating via ".back()", ".forward()", etc.
    """

    def __init__(
        self,
        layout: Callable | Widget | None = None,
        use_pages: bool = False,
        pages: list[Page | ModuleType | str | Callable] | None = None,
        driver_class: type[Driver] | None = None,
        css_path: CSSPathType | None = None,
        watch_css: bool = False,
        css_theme: str | list[str] | None = None,
        css_themes: dict[str, list[CSSPathType]] | None = None,
        logger: logging.Logger | None = None,
    ) -> None:
        """Initialize an application with a widget for the layout.

        Compatibility alias for ExtendedApp.__init__() call with Dash variable names.

        Args:
            layout: Root widget, or function to create root widget.
                Use a callable to delay creation until end of initialization.
                Defaults to a blank Container with "content" id.
            use_pages: Whether to enable multi-page application support.
                When enabled, this will include automatic URL routing callbacks via "widgets.Location".
                Force enabled if "pages" are provided. Enabling without "pages" allows registering pages later.
            pages: Initial pages to load into multi-page applications.
                Refer to "register_page()" for options.
            driver_class: Driver class or `None` to auto-detect.
                This will be used by some Textual tools.
            css_path: Path to CSS or `None` to use the `CSS_PATH` class variable.
                To load multiple CSS files, pass a list of strings or paths which will be loaded in order.
            watch_css: Reload CSS if the files changed.
                This is set automatically if you are using `textual run` with the `dev` switch.
            css_theme: Initial CSS theme to load from "css_themes".
                Themes are applied in addition to base "css_path" values, rather than in place of.
            css_themes: Mapping of CSS paths by string names, or `None` to use the `CSS_THEMES` class variable.
            logger: Custom logger to send callback messages to.

        Raises:
            CssPathError: When the supplied CSS path(s) are an unexpected type.
        """
        super().__init__(
            child=layout,
            use_pages=use_pages,
            pages=pages,
            driver_class=driver_class,
            css_path=css_path,
            watch_css=watch_css,
            css_theme=css_theme,
            css_themes=css_themes,
            logger=logger,
        )

    def callback(
        self,
        *dependencies: Dependency,
    ) -> Callable:
        """Register a callback that triggers when observed values change.

        Compatibility alias for ExtendedApp.when() calls.

        Example:
            @app.callback(
                Input('url', 'path'),
                Output('content-wrapper', 'content'),
            )
            def route_url(path: str) -> str:
                ...

        Args:
            dependencies: Positional arguments containing one or more observation Dependencies.
                May be original observation types (Modified/Select/Update) or aliased types (Input/State/Output).

        Returns:
            Decorator to register a function as an input/output reaction to one or more property changes.
        """
        return super().when(
            *dependencies,
        )

    def register_page(  # Intentional renames for Dash compatibility. pylint: disable=arguments-renamed, too-many-arguments
        self,
        module: Page | ModuleType | str | Callable | None = None,
        path: str | None = None,
        path_template: str | None = None,
        name: str | None = None,
        order: int = 0,
        redirect_from: str | list[str] | None = None,
        layout: Callable | None = None,
    ) -> None:
        """Set up a URL path to provide a layout in a multi-page application.

        The "page_registry" global or app attribute can also be used to create page navigation links
        by application template/layout authors.

        Compatibility alias for ExtendedApp.register_page() calls.

        Args:
            module: A module, or module path, where the remaining page's variables are defined.
                e.g. If calling from within the module itself: "__name__"
                e.g. If calling from another module: "mylib.home"
            path: URL Path, with or without variables. e.g. "/", "/home", "/documents/{document_name}"
                Inferred from the "module" or "layout" if not provided.
                    e.g. "mylib.home" -> "/home"
                    e.g. "home_page" -> "/home_page"
                Variables marked as {variable_name} in paths will be passed to "layout" as keyword arguments.
            path_template: Compatibility alias for "path", no functional difference.
            name: The name of the page link, such as what might be shown in navigation menus.
                Inferred from the "path" if not provided.
                    e.g. "home_page" -> "Home Page"
            order: The relative order to sort pages in the "page_registry", such as ordering in navigation menus.
            redirect_from: Paths that should redirect to this page's path. e.g. "/v1/home"
            layout: Function to call to generate the widget(s) used in the page's layout.
        """
        super().register_page(
            page=module,
            path=path_template or path,
            name=name,
            order=order,
            redirect_from=redirect_from,
            layout=layout,
        )


def Input(  # Treat this function as a class factory. pylint: disable=invalid-name
    id_or_exception: str | SupportsID | type[BaseException],
    property_or_event: str | type | None = None,
) -> InputType:
    """Initialize base dependency identification.

    Args:
        id_or_exception: ID, object with ID, or exception type, that a component uses to trigger callbacks.
        property_or_event: Property name, or event class type, on the component that triggers callbacks.

    Returns:
        Attribute watcher if provided strings, message watcher if string & class, or exception watcher if error type.

    Raises:
        ValueError if any combination of value is incorrect for conversion into an input type dependency
    """
    if isinstance(id_or_exception, str):
        if isinstance(property_or_event, str):
            return Modified(id_or_exception, property_or_event)
        if isinstance(property_or_event, type):
            return Published(id_or_exception, property_or_event)
        raise ValueError("Input dependency second argument can only be an attribute or class")
    if isinstance(id_or_exception, type) and issubclass(id_or_exception, BaseException):
        return Raised(id_or_exception)
    raise ValueError("Input dependency first argument can only be an id to watch or exception to catch")



"""Container widget used for routing paged content."""

from typing import Any
from typing import Iterable

from textual.events import Message
from textual.reactive import reactive

from . import Callback
from . import Container
from . import Widget


class PageContainer(Container):
    """Container used to display page contents, and signal to multi-page applications which ID to use in callbacks.

    If not used in a multi-page app, it is functionally the same as a Container. It also contains attributes and logic
    similar to a ContentSwitcher.
    """

    page: str | None = reactive(None, repaint=False, init=False)

    def __init__(
        self,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize the paged container with initial children and additional tracking.

        Args:
            *children: Child widgets.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )
        self._page_cache: dict[str, Widget] = {}

    async def add_page(self, page: str, content: Widget) -> None:
        """Add a page to the cache.

        Args:
            page: Name of the page to use during cache requests.
            content: Content to display when the current page is updated.
        """
        with self.app.batch_update():
            if page in self._page_cache:
                await self._page_cache.pop(page).remove()
            content.display = False
            await self.mount(content)
        self._page_cache[page] = content
        if not self.page:
            self.page = page

    def is_cached(self, page: str) -> bool:
        """Check whether the container has the request page in the cache.

        Return:
            True if the page is found in the cache, False otherwise.
        """
        return page in self._page_cache

    async def remove_page(self, page: str) -> None:
        """Remove a page from the cache."""
        with self.app.batch_update():
            if page in self._page_cache:
                await self._page_cache.pop(page).remove()
        if self.page == page:
            self.page = None

    def watch_page(self, old: str | None, new: str | None) -> None:
        """React to the current visible child choice being changed.

        Args:
            old: The old path that was displayed.
            new: The new path to be shown.
        """
        with self.app.batch_update():
            if old_widget := self._page_cache.get(old):
                old_widget.display = False
            if new_widget := self._page_cache.get(new):
                new_widget.display = True



"""Extended Textual ProgressBar widget."""

from typing import Any
from typing import Iterable

from textual import events
from textual import widgets

from .._extensions import Callback
from .._extensions import WidgetExtension


class ProgressBar(widgets.ProgressBar, WidgetExtension):
    """An extended progress bar widget."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        total: float | None = None,
        *,
        show_bar: bool = True,
        show_percentage: bool = True,
        show_eta: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize a Progress Bar widget.

        Args:
            total: The total number of steps in the progress if known.
            show_bar: Whether to show the bar portion of the progress bar.
            show_percentage: Whether to show the percentage status of the bar.
            show_eta: Whether to show the ETA countdown of the progress bar.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            total=total,
            show_bar=show_bar,
            show_percentage=show_percentage,
            show_eta=show_eta,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Extended Textual vertical list view widget."""

from __future__ import annotations

from typing import Any
from typing import Iterable

from textual import events
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import ListItem
from textual.widgets import ListView as TextualListView
from typing_extensions import override

from ._extensions import Callback
from ._extensions import WidgetExtension
from ._list_item_header import ListItemHeader


class ListView(TextualListView, WidgetExtension):
    """An extended vertical list view widget."""

    default_disabled_messages = [
        # Disable click events, they are already handled by _on_list_item__child_clicked.
        events.Click,
    ]

    # Most recently highlighted item in the list.
    highlighted: ListItem | None = reactive(None, repaint=False, init=False)

    def __init__(
        self,
        *children: ListItem,
        initial_index: int | None = 0,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        auto_highlight: bool = True,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize a ListView with extension arguments.

        Args:
            *children: The ListItems to display in the list.
            initial_index: The index that should be highlighted when the list is first mounted.
            name: The name of the widget.
            id: The unique ID of the widget used in CSS/query selection.
            classes: The CSS classes of the widget.
            disabled: Whether the ListView is disabled or not.
            auto_highlight: Whether the ListView automatically highlights the first item on focus.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            *children,
            initial_index=initial_index,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )
        self.auto_highlight = auto_highlight

    def on_focus(self, event: events.Focus) -> None:
        """Automatically highlight the first item in the list when the list is focused."""
        super()._on_focus(event)
        if self.auto_highlight and self.has_focus and self.index is None and self._nodes:
            self.index = 0

    def _on_list_item__child_clicked(
        self,
        event: ListItem._ChildClicked,
    ) -> None:  # pylint: disable=protected-access
        """Stop ListView child click events from propagating past ListView.

        Absorb the event propagation at the ListView level, do not send up the stack, but call default.
        The default operation will send a Selected event that observers can handle.
        """
        # Absorb the event propagation at the ListView level, do not send up the stack, but call default.
        event.stop()
        event.prevent_default()
        super()._on_list_item__child_clicked(event)

    def on_list_view_highlighted(self, _: ListView.Highlighted) -> None:
        """Update reactive attributes on highlight changes."""
        self.highlighted = self.highlighted_child

    @override
    async def _replace(
        self,
        widgets: list[ListItem],
    ) -> None:
        self.index = None
        await super()._replace(widgets)

    def watch_index(self, old_index: int, new_index: int) -> None:
        """Updates the highlighted when the index changes.

        Overrides ListView watch_index to allow support detecting header items that should be skip highlighting.
        """
        if self._is_valid_index(old_index):
            old_child = self._nodes[old_index]
            old_child.highlighted = False

        new_child: Widget | None
        if self._is_valid_index(new_index):
            new_child = self._nodes[new_index]
            if isinstance(new_child, ListItemHeader):
                if new_index == 0:
                    if old_index == 0 or not self._nodes:
                        self.index = None
                    else:
                        self.index = new_index + 1
                else:
                    self.index = new_index + 1 if new_index > old_index else new_index - 1
                return
            new_child.highlighted = True
        else:
            new_child = None

        self._scroll_highlighted_region()
        self.post_message(self.Highlighted(self, new_child))



"""Utilities for configuring and managing multi-page applications."""

from __future__ import annotations

import importlib
from types import ModuleType
from typing import Callable

_GLOBAL_PAGE_MAP: dict[str, Page] = {}


class Page:
    """Configuration for a page in a multi-page application routed via URL path."""

    def __init__(
        self,
        layout: Callable,
        path: str | None = None,
        name: str | None = None,
        order: int = 0,
        redirect_from: str | list[str] | None = None,
    ) -> None:
        """Initialize the page configuration.

        Args:
            layout: Function to call to generate the widget(s) used in the page's layout.
            path: URL Path, with or without variables. e.g. "/", "/home", "/documents/{document_name}"
                Inferred from the "layout" if not provided.
                    e.g. "home_page" -> "/home_page"
                Variables marked as {variable_name} in paths will be passed to "layout" as keyword arguments.
            name: The name of the page link, such as what might be shown in navigation menus.
                Inferred from the "path" if not provided.
                    e.g. "home_page" -> "Home Page"
            order: The relative order to sort pages in the "page_registry", such as ordering in navigation menus.
            redirect_from: Paths that should redirect to this page's path. e.g. "/v1/home"
        """
        if not path:
            path = layout.__name__.removeprefix("layout_")
        self.layout = layout
        self.path = path if path.startswith("/") else f"/{path}"
        self.name = name or self.path.strip("/").replace("_", " ").title()
        self.order = order
        self.redirect_from = [redirect_from] if isinstance(redirect_from, str) else redirect_from

    @staticmethod
    def from_module(module: ModuleType | str) -> Page:
        """Convert a module into an application page.

        Args:
            module: Loaded module, or path to a module, with page attributes such as "layout".

        Returns:
            Newly created page with attributes, such as path, populated by the module's attributes and variables.
        """
        kwargs = {}
        vars_to_save = ("layout", "path", "name", "order", "redirect_from")
        if isinstance(module, str):
            module = importlib.import_module(module)
        for var_name in dir(module):
            if var_name not in vars_to_save:
                continue
            var = getattr(module, var_name)
            kwargs[var_name] = var

        if "layout" not in kwargs:
            raise ValueError(f"Module {module} must contain a layout function to be used as a page")
        if "path" not in kwargs:
            kwargs["path"] = module.__name__.rsplit(".", 1)[-1]
        return Page(**kwargs)

    @staticmethod
    def to_page(page: Page | ModuleType | str | Callable) -> Page:
        """Convert a page compatible object into a full application page.

        Args:
            page: Object compatible with conversion into an application page, such as module, module name, for function.
                e.g. "modules" will create full pages after searching for all values compatible with Page.
                e.g. "functions" will only create a basic Page with the path based on name of the layout function.

        Returns:
            Newly created page with attributes, such as path, populated based on the type of object provided.
        """
        if isinstance(page, (ModuleType, str)):
            page = Page.from_module(page)
        elif isinstance(page, Callable):
            page = Page(page)
        return page


def register_page(  # pylint: disable=too-many-arguments
    page: Page | ModuleType | str | Callable | None = None,
    path: str | None = None,
    name: str | None = None,
    order: int = 0,
    redirect_from: str | list[str] | None = None,
    layout: Callable | None = None,
    page_map: dict[str, Page] | None = None,
) -> Page:
    """Register a URL path to a layout shared across all multi-page applications.

    Args:
        page: A module, or module path, where the remaining page's variables are defined.
            e.g. If calling from within the module itself: "__name__"
            e.g. If calling from another module: "mylib.home"
        path: URL Path, with or without variables. e.g. "/", "/home", "/documents/{document_name}"
            Inferred from the "module" or "layout" if not provided.
                e.g. "mylib.home" -> "/home"
                e.g. "layout_home_page" -> "/home_page"
            Variables marked as {variable_name} in paths will be passed to "layout" as keyword arguments.
        name: The name of the page link, such as what might be shown in navigation menus.
            Inferred from the "path" if not provided.
                e.g. "home_page" -> "Home Page"
        order: The relative order to sort pages in the "page_registry", such as ordering in navigation menus.
        redirect_from: Paths that should redirect to this page's path. e.g. "/v1/home"
        layout: Function to call to generate the widget(s) used in the page's layout.
        page_map: All currently registered pages for a multi-page app.
            Modified in place when page is registered. Defaults to global pages shared across apps.

    Returns:
        Final page registered to the path, and created from arguments if needed.
    """
    if not page and not layout:
        raise ValueError("Either a module or a layout must be provided to create a page")
    if isinstance(redirect_from, str):
        redirect_from = [redirect_from]
    if page is not None:
        page = Page.to_page(page)
        if path:
            page.path = path
        if redirect_from:
            page.redirect_from = redirect_from
        if layout:
            page.layout = layout
        if name:
            page.name = name
        if order:
            page.order = order
    else:
        page = Page(
            layout=layout,
            path=path,
            name=name,
            order=order,
            redirect_from=redirect_from,
        )
    paths = [page.path]
    if page.redirect_from:
        paths.extend(page.redirect_from)
    page_map = page_map if page_map is not None else _GLOBAL_PAGE_MAP
    for page_path in paths:
        if page.path in page_map:
            raise ValueError(f"Duplicate page found for path: {page_path}")
        page_map[page_path] = page
    return page



"""Widget container for basic data storage and sharing between callbacks."""

from __future__ import annotations

import time
from typing import Any
from typing import Iterable
from typing import TypeVar

from textual import events
from textual.reactive import reactive

from ._extensions import Callback
from ._extensions import Widget

JsonType = TypeVar("JsonType", dict, list, bool, float, int, str, None)  # Custom type. pylint: disable=invalid-name


class Store(Widget):
    """Hidden widget for basic data storage.

    Intended to only be used for sharing data between callbacks.
    """

    DEFAULT_CSS = """
    Store {
        display: none;
        width: 0;
        height: 0;
    }
    """

    # The currently stored data.
    data: JsonType = reactive(None, repaint=False, init=False)
    # The timestamp from the last time the value was modified.
    modified_timestamp: float = reactive(-1.0, repaint=False, init=False)
    # Sentinel value used to trigger clears from callbacks. Set to True to manually trigger a clear.
    clear_data: bool = reactive(False, always_update=True, repaint=False, init=False)

    class Updated(events.Message):
        """Posted when the backend data is updated."""

        def __init__(self, store: Store, data: Any, modified_timestamp: float) -> None:
            """Initialize the data update event.

            Args:
                store: The store that sent the event.
                data: The new data stored.
                modified_timestamp: When the data was modified.
            """
            super().__init__()
            self.store = store
            self.data = data
            self.modified = modified_timestamp

        @property
        def control(self) -> Store:
            """The store that contains the data."""
            return self.store

    def __init__(
        self,
        data: Any = None,
        id: str | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize the data store.

        Args:
            data: Initial data to store.
            id: The ID of the widget in the DOM.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(id=id, disabled_messages=disabled_messages, callbacks=callbacks)
        self.data = data

    def watch_clear_data(self, new_value: bool) -> None:
        """Monitor the sentinel attribute for clear data requests from callbacks.

        Args:
            new_value: New clear data value set on the reactive attribute.
        """
        if new_value:
            self.data = None

    def watch_data(self, _: JsonType) -> None:
        """Monitor the data in order to update the modified timestamp."""
        self.modified_timestamp = time.time()
        self.post_message(self.Updated(self, self.data, self.modified_timestamp))



"""Multiple choice dropdown widget."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import Iterable

from rich.console import RenderableType
from rich.text import Text
from rich.text import TextType
from textual import events
from textual import on
from textual.css.query import NoMatches
from textual.message import Message
from textual.reactive import var
from textual.types import NoSelection
from textual.types import SelectType
from textual.widgets import Select
from textual.widgets import SelectionList
from textual.widgets._select import SelectCurrent
from textual.widgets.select import EmptySelectError
from textual.widgets.select import InvalidSelectValueError
from textual.widgets.selection_list import Selection
from textual.widgets.selection_list import SelectionType

from ._extensions import Callback
from ._textual._containers import Vertical

if TYPE_CHECKING:
    from textual.app import ComposeResult

BLANK = Select.BLANK


class MultiSelectOverlay(SelectionList):  # pylint: disable=too-many-ancestors
    """The 'pop-up' overlay for the MultiSelect control.

    Based off textual.widgets.select.SelectOverlay, but uses SelectionList for multiple choice selection.
    """

    BINDINGS = [
        ("escape", "dismiss"),
        ("backspace,delete", "clear"),
    ]

    DEFAULT_CSS = """
    MultiSelectOverlay {
        border: tall $background;
        background: $panel;
        color: $text;
        width: 100%;
        padding: 0 1;
    }
    MultiSelectOverlay:focus {
        border: tall $background;
    }
    MultiSelectOverlay > .option-list--option {
        padding: 0 1;
    }
    """

    allow_blank: bool = var[bool](False, init=False)

    @dataclass
    class Dismiss(Message):
        """Inform ancestor the overlay should be dismissed."""

        lost_focus: bool = False

    @dataclass
    class UpdateSelection(Message):
        """Inform ancestor the selection changed."""

        selected: list[SelectType]

    def _action_clear(self) -> None:
        """Clear the selected item."""
        self.deselect_all()

    def _action_dismiss(self) -> None:
        """Dismiss the overlay."""
        self.post_message(self.Dismiss())

    def _on_blur(self, _event: events.Blur) -> None:
        """Dismiss the overlay on blur."""
        self.post_message(self.Dismiss(lost_focus=True))

    def _on_selection_list_selected_changed(self, event: SelectionList.SelectedChanged) -> None:
        """Inform parent when an option is selected."""
        event.stop()
        self.post_message(self.UpdateSelection(event.selection_list.selected))

    def _deselect(self, value: SelectionType) -> bool:
        """Mark the given selection as not selected.

        Override default behavior to disable deselection if `allow_blank` is false, and it would deselect last value.
        """
        if len(self.selected) <= 1 and not self.allow_blank:
            return False
        return super()._deselect(value)


class MultiSelect(Generic[SelectType], Vertical, can_focus=True):  # pylint: disable=too-many-ancestors
    """Widget to select multiple choices from a list of possible options.

    A MultiSelect displays the current selection(s) by default. When activated, the widget displays an overlay
    with a list of all possible options.

    Based off textual.widgets.select.Select, but uses SelectionList for multiple choice selection instead of OptionList.
    """

    BINDINGS = [("enter,down,space,up", "show_overlay")]
    DEFAULT_CSS = """
    MultiSelect {
        height: auto;

        & > MultiSelectOverlay {
            width: 1fr;
            display: none;
            height: auto;
            max-height: 12;
            overlay: screen;
            constrain: y;
        }

        &:focus > SelectCurrent {
            border: tall $accent;
        }

        .up-arrow {
            display: none;
        }

        &.-expanded .down-arrow {
            display: none;
        }

        &.-expanded .up-arrow {
            display: block;
        }

        &.-expanded > MultiSelectOverlay {
            display: block;
        }

        &.-expanded > SelectCurrent {
            border: tall $accent;
        }
    }
    """

    expanded: bool = var[bool](False, init=False)
    prompt: str = var[str]("Select", init=False)
    values: list[SelectType] | NoSelection = var[list[SelectType] | NoSelection](BLANK, init=False)

    class Changed(Message):
        """Posted when the selected values change."""

        def __init__(self, select: MultiSelect[SelectType], selected: list[SelectType] | NoSelection) -> None:
            """Initialize the Changed message."""
            super().__init__()
            self.select = select
            self.selected = selected

        @property
        def control(self) -> MultiSelect[SelectType]:
            """The MultiSelect that sent the message."""
            return self.select

    def __init__(
        self,
        options: Iterable[
            Selection[SelectionType] | tuple[TextType, SelectionType] | tuple[TextType, SelectionType, bool]
        ],
        *,
        prompt: str = "Select",
        allow_blank: bool = True,
        values: list[SelectType] | NoSelection = BLANK,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize the MultiSelect control.

        Args:
            options: Options to select from. If no options are provided then
                `allow_blank` must be set to `True`.
            prompt: Text to show in the control when no option is selected.
            allow_blank: Enables or disables the ability to have the widget in a state
                with no selection made, in which case its value is set to the constant
                [`Select.BLANK`][textual.widgets.Select.BLANK].
            values: Initial value(s) selected. Should be one or mor of the values in `options`.
                If no initial value is set and `allow_blank` is `False`, the widget
                will auto-select the first available option.
            name: The name of the select control.
            id: The ID of the control in the DOM.
            classes: The CSS classes of the control.
            disabled: Whether the control is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.

        Raises:
            EmptySelectError: If no options are provided and `allow_blank` is `False`.
        """
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )
        self._allow_blank = allow_blank
        self.prompt = prompt
        self._values = values
        self._options = []
        self._select_options = []
        self._setup_variables_for_options(options)

    def action_show_overlay(self) -> None:
        """Show the overlay."""
        select_current = self.query_one(SelectCurrent)
        select_current.has_value = bool(self.values) and self.values != BLANK
        self.expanded = True
        # If we haven't opened the overlay yet, highlight the first option.
        select_overlay = self.query_one(MultiSelectOverlay)
        if select_overlay.highlighted is None:
            select_overlay.action_first()

    def clear(self) -> None:
        """Clear the selection if `allow_blank` is `True`.

        Raises:
            InvalidSelectValueError: If `allow_blank` is set to `False`.
        """
        try:
            self.values = BLANK
        except InvalidSelectValueError:
            raise InvalidSelectValueError("Can't clear selection if allow_blank is set to False.") from None

    def compose(self) -> ComposeResult:
        """Compose MultiSelect with overlay and current value."""
        yield SelectCurrent(self.prompt)
        multiselect = MultiSelectOverlay()
        multiselect.allow_blank = self._allow_blank
        yield multiselect

    def _init_selected_options(self, hints: list[SelectType] | NoSelection = BLANK) -> None:
        """Initializes the selected option for the `MultiSelect`."""
        if hints == BLANK and not self._allow_blank:
            hints = [self._options[0][1]]
        self.values = hints

    def is_blank(self) -> bool:
        """Indicates whether this `MultiSelect` is blank or not.

        Returns:
            True if the selection is blank, False otherwise.
        """
        return self.values == BLANK

    def _on_mount(self, _: events.Mount) -> None:
        """Set initial values."""
        self._setup_options_renderables()
        self._init_selected_options(self._values)

    @on(SelectCurrent.Toggle)
    def _select_current_toggle(self, event: SelectCurrent.Toggle) -> None:
        """Show the overlay when toggled."""
        event.stop()
        self.expanded = not self.expanded

    @on(MultiSelectOverlay.Dismiss)
    def _select_overlay_dismiss(self, event: MultiSelectOverlay.Dismiss) -> None:
        """Dismiss the overlay."""
        event.stop()
        self.expanded = False
        if not event.lost_focus:
            # If the overlay didn't lose focus, don't to re-focus the select.
            self.focus()

    def set_options(
        self,
        options: Iterable[
            Selection[SelectionType] | tuple[TextType, SelectionType] | tuple[TextType, SelectionType, bool]
        ],
    ) -> None:
        """Set the options for the MultiSelect.

        This will reset the selection. The selection will be empty, if allowed, otherwise
        the first valid option is picked.

        Args:
            options: An iterable containing the renderable to display for each option (optional),
                the corresponding internal value (required), and the initial state of the selection (optional).

        Raises:
            EmptySelectError: If the options iterable is empty and `allow_blank` is `False`.
        """
        self._setup_variables_for_options(options)
        self._setup_options_renderables()
        self._init_selected_options()

    def _setup_options_renderables(self) -> None:
        """Sets up the `Selection` renderables associated with the `MultiSelect` options."""
        self._select_options: list[Selection] = [
            Selection(prompt, value, initial_state=initial_state) for prompt, value, initial_state in self._options
        ]
        selection_list = self.query_one(MultiSelectOverlay)
        selection_list.clear_options()
        for option in self._select_options:
            selection_list.add_option(option)

    def _setup_variables_for_options(
        self,
        options: Iterable[
            Selection[SelectionType] | tuple[TextType, SelectionType] | tuple[TextType, SelectionType, bool]
        ],
    ) -> None:
        """Setup function for the auxiliary variables related to options.

        This method sets up `self._options` and `self._legal_values`.
        """
        self._options: list[tuple[RenderableType, SelectType, bool]] = []
        for option in options:
            if not isinstance(option, tuple):
                option = (str(option), option, False)
            if not len(option) == 3:
                option = option + (False,)
            self._options.append(option)

        if not self._options and not self._allow_blank:
            raise EmptySelectError("MultiSelect options cannot be empty if selection can't be blank.")

        self._legal_values: set[SelectType | NoSelection] = {value for _, value, __ in self._options}
        if self._allow_blank:
            self._legal_values.add(BLANK)

    @on(MultiSelectOverlay.UpdateSelection)
    def _update_selection(self, event: MultiSelectOverlay.UpdateSelection) -> None:
        """Update the current selection."""
        event.stop()
        values = event.selected
        if values != self.values:
            self.values = values
            self.post_message(self.Changed(self, values))

    def _validate_values(
        self,
        values: list[SelectType] | NoSelection,
    ) -> list[SelectType] | NoSelection:
        """Ensure the new values only contain valid options.

        If `allow_blank` is `True`, `None` is also a valid value and corresponds to no selection.

        Raises:
            InvalidSelectValueError: If the new value does not correspond to any known value.
        """
        if not self._allow_blank and (not values or values == BLANK):
            raise InvalidSelectValueError("Can't clear selection if allow_blank is set to False.")
        new_values = []
        if values == BLANK:
            # Convert to a list for consistent check logic.
            new_values.append(BLANK)
        for sub_value in new_values:
            if sub_value not in self._legal_values:
                # It would make sense to use `None` to flag that the Select has no selection,
                # so we provide a helpful message to catch this mistake in case people didn't
                # realise we use a special value to flag "no selection".
                help_text = " Did you mean to use MultiSelect.clear()?" if sub_value is None else ""
                raise InvalidSelectValueError(f"Illegal multiselect value {sub_value!r}.{help_text}")
        return values

    def _watch_expanded(self, expanded: bool) -> None:
        """Display or hide overlay."""
        overlay = self.query_one(MultiSelectOverlay)
        self.set_class(expanded, "-expanded")
        if expanded:
            overlay.focus()
            if self.values == BLANK:
                overlay.deselect_all()
                self.query_one(SelectCurrent).has_value = False
            else:
                for _, value, __ in self._options:
                    if value in self.values:
                        overlay.select(value)
                self.query_one(SelectCurrent).has_value = True

    def _watch_prompt(self, prompt: str) -> None:
        """Update the current prompt when it changes."""
        if not self.is_mounted:
            return
        select_current = self.query_one(SelectCurrent)
        select_current.placeholder = prompt
        if not self._allow_blank:
            return
        if self.values == BLANK:
            select_current.update(BLANK)
        option_list = self.query_one(MultiSelectOverlay)
        option_list.replace_option_prompt_at_index(0, Text(prompt, style="dim"))

    def _watch_values(self, values: list[SelectType] | NoSelection) -> None:
        """Update the current value when it changes."""
        self._values = values if values else BLANK
        try:
            select_current = self.query_one(SelectCurrent)
        except NoMatches:
            pass
        else:
            if self.values == BLANK:
                select_current.update(BLANK)
            else:
                prompts = []
                for prompt, option_value, _ in self._options:
                    if option_value in values:
                        prompts.append(prompt)
                select_current.update(", ".join(prompts) or BLANK)



"""Extended Textual Footer widget."""

from typing import Any
from typing import Iterable

from textual import events
from textual import widgets

from .._extensions import Callback
from .._extensions import WidgetExtension


class Footer(widgets.Footer, WidgetExtension):
    """An extended simple footer widget which docks itself to the bottom of the parent container."""

    def __init__(
        self,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize the footer widget.

        Args:
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__()
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Extended Textual Tab widgets."""

from typing import Any
from typing import Iterable

from rich.text import TextType
from textual import events
from textual import widgets

from .._extensions import Callback
from .._extensions import WidgetExtension


class Tab(widgets.Tab, WidgetExtension):
    """An extended widget to manage a single tab within a Tabs widget."""

    def __init__(
        self,
        label: TextType,
        *,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize a Tab.

        Args:
            label: The label to use in the tab.
            id: Optional ID for the widget.
            classes: Space separated list of class names.
            disabled: Whether the tab is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            label,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )


class Tabs(widgets.Tabs, WidgetExtension):
    """An extended row of tabs."""

    def __init__(
        self,
        *tabs: Tab | TextType,
        active: str | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize a Tabs widget.

        Args:
            *tabs: Positional argument should be explicit Tab objects, or a str or Text.
            active: ID of the tab which should be active on start.
            name: Optional name for the input widget.
            id: Optional ID for the widget.
            classes: Optional initial classes for the widget.
            disabled: Whether the input is disabled or not.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        super().__init__(
            *tabs,
            active=active,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )



"""Progressive horizontal set of ListViews with tracking for active item and peeking at next menu."""

from __future__ import annotations

from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Iterable

from textual import containers
from textual import events
from textual.await_complete import AwaitComplete
from textual.binding import Binding
from textual.binding import BindingType
from textual.reactive import reactive
from textual.widget import Widget

from ._extensions import Callback
from ._extensions import WidgetExtension
from ._extensions import walk_all_children
from ._list_item import ListItem
from ._list_item_header import ListItemHeader
from ._list_item_meta import ListItemMeta
from ._list_view import ListView


class HorizontalMenus(containers.HorizontalScroll, WidgetExtension):
    """Progressive horizontal set of ListViews with tracking for active item(s).

    Compared to a Tree view, horizontal menus provide quick navigation and visibility into nested objects.
    The menus can be quickly focused, selected, etc., with less vertical traversal in large objects.
    "Previewing" the next menu in a set of menus is also allowed.
    """

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("left", "focus_previous", show=False),
        Binding("right", "focus_next", show=False),
    ]

    # Currently focused item across sub-menus.
    focused: ListItem | None = reactive(None, repaint=False, init=False)
    # Current primary item highlighted across sub-menus (right-most).
    highlighted: ListItem | None = reactive(None, repaint=False, init=False)
    # All highlighted items across sub-menus.
    highlights: list[ListItem] = reactive([], repaint=False, init=False)

    class Focused(events.Message, bubble=True):
        """Posted when the focused item changes."""

        def __init__(self, horizontal_menu: HorizontalMenus, item: ListItem | None) -> None:
            """Initialize focused event.

            Args:
                horizontal_menu: Parent menu containing the item.
                item: New item that was focused.
            """
            super().__init__()
            self.horizontal_menu: HorizontalMenus = horizontal_menu
            self.item: ListItem | None = item

        @property
        def control(self) -> HorizontalMenus:
            """The primary controlling widget that contains the focused item, used by "on()" decorator."""
            return self.horizontal_menu

    class Highlighted(events.Message, bubble=True):
        """Posted when the highlighted item changes."""

        def __init__(self, horizontal_menu: HorizontalMenus, item: ListItem | None) -> None:
            """Initialize highlight event.

            Args:
                horizontal_menu: Parent menu containing the item.
                item: New item that was highlighted.
            """
            super().__init__()
            self.horizontal_menu: HorizontalMenus = horizontal_menu
            self.item: ListItem | None = item

        @property
        def control(self) -> HorizontalMenus:
            """The primary controlling widget that contains the highlighted item, used by "on()" decorator."""
            return self.horizontal_menu

    def __init__(
        self,
        *children: Widget | ListItem | ListItemMeta,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
        can_focus: bool = False,
        menu_creator: Callable[[int, list[ListItem]], Widget] | None = None,
        styles: dict[str, Any] | None = None,
        disabled_messages: Iterable[type[events.Message]] | None = None,
        callbacks: dict[str, Callback] | None = None,
    ) -> None:
        """Initialize horizontal menus.

        Args:
            *children: Initial menus, or menu items, to display.
                Each "menu" widget must contain a ListView, but does not have to be a ListView itself.
                If children are ListItem or ListItemMeta objects, an initial ListView will be made with menu_creator.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
            can_focus: Whether the parent widget can be focused, or only the children widgets.
                If enabled, it will take focus after all the nested ListViews, but before the next sibling.
            menu_creator: Called to create new sub-menus when an item with children is highlighted.
            styles: Local inline styles to apply on top of the class' styles for only this instance.
            disabled_messages: List of messages to disable on this widget instance only.
            callbacks: Mapping of callbacks to send messages to instead of sending to default handler.
        """
        self.menu_creator = menu_creator or self._default_menu_creator
        if any(isinstance(child, (ListItem, ListItemMeta)) for child in children):
            if not all(isinstance(child, (ListItem, ListItemMeta)) for child in children):
                raise ValueError("All initial children must be of same type: ListItem, ListItemMeta, or ListView")
            self._update_menu_index(children, 0)
            children = [
                self.menu_creator(
                    0,
                    [child.to_item() if isinstance(child, ListItemMeta) else child for child in children],
                )
            ]
        super().__init__(
            *children,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.__extend_widget__(
            styles=styles,
            disabled_messages=disabled_messages,
            callbacks=callbacks,
        )
        self.can_focus = can_focus
        self.menus = []
        for child in children or []:
            menu = None
            if isinstance(child, ListView):
                menu = child
            else:
                for node in walk_all_children(child):
                    if isinstance(node, ListView):
                        menu = node
                        break
            if menu is None:
                raise ValueError("All menu children must contain a ListView widget")
            self.menus.append(menu)

    def action_focus_next(self) -> None:
        """Focus the next widget and ensure the first item is highlighted."""
        self.app.action_focus_next()
        focused = self.app.focused
        if focused in self.menus and focused.index is None and any(focused.children):
            focused.index = 0

    def _add_menu(self, items: list[ListItem]) -> AwaitComplete:
        """Create and add a menu to the list of available menus."""
        new_menu = self.menu_creator(len(self.menus), items)
        mount = None
        if new_menu is not None:
            list_view = None
            if isinstance(new_menu, ListView):
                list_view = new_menu
            else:
                for node in walk_all_children(new_menu):
                    if isinstance(node, ListView):
                        list_view = node
                        break
            if list_view is None:
                raise ValueError("Menus must contain a ListView widget for navigation")
            self.menus.append(list_view)
            mount = self.mount(new_menu)
            await_complete = AwaitComplete(mount)
        else:
            await_complete = AwaitComplete()
        self.call_next(await_complete)
        return await_complete

    @staticmethod
    def _default_menu_creator(menu_index: int, items: list) -> ListView | None:
        """Default menu factory to create submenus when another submenu highlight changes."""
        return (
            ListView(
                *items,
                auto_highlight=False,
                initial_index=None,
                classes=f"--horizontal-menu-{menu_index}",
            )
            if any(not isinstance(item, ListItemHeader) for item in items)
            else None
        )

    def _find_highlighted(self) -> ListItem | None:
        """Find the rightmost highlighted (most recent) item across the menu hierarchy."""
        for menu in reversed(self.menus):
            highlighted = menu.highlighted_child
            if highlighted:
                return highlighted
        return None

    def _find_highlights(self) -> list[ListItem]:
        """Find the highlighted items across all the menus in the hierarchy."""
        highlights = []
        for menu in reversed(self.menus):
            highlighted = menu.highlighted
            if highlighted:
                highlights.append(highlighted)
        return highlights

    def on_descendant_focus(self, focus_event: events.DescendantFocus) -> None:
        """Update highlighted and focused items when focus on nested menus updates."""
        focused = None
        for menu in reversed(self.menus):
            if not menu.has_focus:
                continue
            highlighted_child = menu.highlighted_child
            if highlighted_child:
                self.focused = highlighted_child
                break
        self.focused = focused
        self.highlighted = self._find_highlighted()
        self.highlights = self._find_highlights()
        if not self.highlighted and self.menus and focus_event.widget == self.menus[0]:
            focus_event.widget.index = 0

    def on_list_view_highlighted(self, event: ListView.Highlighted) -> None:
        """Update highlighted items and focused menus when highlight on nested item updates."""
        # Ensure ListView with newly highlighted item is focused to ensure no accidental auto highlight on
        # removal of menus > 1 away. Example: Do not auto highlight menu 1 when menu 0 was clicked from menu 2.
        # Auto highlighting occurs in ListView on_focus() to ensure that an item is always selected once focused.
        if event.item and event.item.highlighted:
            self.screen.set_focus(event.list_view)
            self.focused = event.item
            if event.item not in self.highlights:
                self.highlighted = event.item
            self.highlights = self._find_highlights()
        event.stop()
        event.prevent_default()

    def remove_menus(self, last_index: int) -> None:
        """Remove all menus after a specific index.

        Args:
            last_index: Index of the last menu to show.
        """
        for index, child in reversed(list(enumerate(self.children))):
            if index <= last_index:
                continue
            child.remove()
            self.menus.pop(index)

    def show_menu(self, index: int, items: list[ListItem]) -> AwaitComplete:
        """Show a new set of items, either by creating a new menu, or updating an existing menu.

        Args:
            index: Position of the menu to create or update.
            items: New items to show in the menu at the provided index.

        Returns:
            An awaitable object that waits for menus to be added or updated.
        """
        self._update_menu_index(items, index)
        if index >= len(self.menus):
            return self._add_menu(items)
        return self._update_menu(index, items)

    def _update_menu(self, index: int, new_items: list[ListItem]) -> AwaitComplete:
        """Update the items in an existing menu."""
        # Do not count headers towards the available items to display in a sub-menu.
        non_headers = [item for item in new_items if not isinstance(item, ListItemHeader)]
        self.remove_menus(index if non_headers else index - 1)
        if index < len(self.menus):
            awaitable = self.menus[index].replace(new_items)
        else:
            awaitable = AwaitComplete()
            self.call_next(awaitable)
        return awaitable

    @staticmethod
    def _update_menu_index(items: Iterable[ListItem | ListItemMeta], index: int) -> None:
        """Update the menu index value on all items in a list."""
        for item in items:
            item.menu_index = index

    def watch_focused(self, old_value: ListItem | None, new_value: ListItem | None) -> None:
        """Monitor the focused item to update listeners.

        Args:
            old_value: Previously focused list item.
            new_value: Newly focused list item.
        """
        if new_value == old_value or not new_value:
            return
        self.post_message(self.Focused(self, new_value))

    def watch_highlighted(self, old_value: ListItem | None, new_value: ListItem | None) -> None:
        """Monitor the highlighted item to update the submenus and listeners.

        Args:
            old_value: Previously highlighted list item.
            new_value: Newly highlighted list item.
        """
        if new_value == old_value or not new_value:
            return
        menu_items = new_value.data.get("menu_items")
        menu_index = new_value.menu_index or 0
        if menu_items:
            list_items = [item.to_item() for item in menu_items]
            self.show_menu(menu_index + 1, list_items)
        else:
            self.remove_menus(menu_index)
        self.post_message(self.Highlighted(self, new_value))
